//----------------------------------------------------------------------------

//   NEC

//-----------------------  : ------------------------------

//     NEC.
//   :  INT0   
//   Timer2.

//    - ST_IDLE.   
//  (AGC pulse)  .  
// INT0   ,   ,
//   Power Down.      ,
//  .

//   ,   
//     INT0.     
// ST_WAIT,   , INT0   
// .      T_DATA  
//  ,    T_REPEAT -  .
//      Timer2.   INT0
//    ,    .
//..        .

//   ST_WAIT    , 
//      .
//   T_DATA,     ,
//     ST_DATA,    .

//        
//  .    T_ONE,   ,
// T_ZERO -  .      RxCode.
//   NEC_BITS (  32),   .

//32-  Code  union   NECStruct,
//    4 : NECpCust, NECnCust, NECpData  NECnData.

//   NEC  NECpCust  NECnCust  
//     (Custom code).   
//   NEC,     
// 16-  .    
//   ,      .

//  NECpData  NECnData     
// .      .   
//,  NECpData    Command,  
//NECpCust  NECnCust    System.   
//  ST_DONE.

//     .   
//,       ST_WAIT.  
// T_REPEAT      
//,    .   Command  System 
//   RxCode.      ST_DONE.

//      ,  
//     .  
//  ,    . 
//,     ST_IDLE.

//         ,
//   32    Timer2,  
//.         
//.        
//:    ST_IDLE,   ,
//INT0      ,   
//Power Down.

//        GetCom().
//  ,     ST_DONE (  ),
//   .     SysUsed, 
//   .    ,
//   .    REM_NO.
// GetCom()    ST_DONE  ST_IDLE.

//----------------------------------------------------------------------------

#include "main.hpp"
#include "remote.hpp"
#include "keyboard.hpp"

//----------------------    : ------------------------

#include "remcodes_rc15a.hpp"

//----------------------    NEC: ------------------------

#define NEC_BITS 32 //NEC protocol bits count
#define T_NEC 1.125 //NEC protocol timebase, ms
#define T_START  (8 * T_NEC) //AGC pulse with 9 ms
#define T_SPACE  (4 * T_NEC) //space before data bits 4.5 ms
#define T_ONE    (2 * T_NEC) //one data bit 2.25 ms (1.5...3.0 ms)
#define T_ZERO   (1 * T_NEC) //zero data bit 1.125 ms (0.7...1.5 ms)
#define T_DATA   (T_START + T_SPACE) //first data pulse delay 13.5 ms (12...15 ms)
#define T_REPEAT (T_START + T_ONE)   //repeat pulse delay 11.25 ms (10..12 ms)

#define T_ZER_MIN 0.7 //zero data bit min time, ms
#define T_ZER_MAX 1.5 //zero data bit max time, ms
#define T_ONE_MIN 1.5 //one data bit min time, ms
#define T_ONE_MAX 3.0 //one data bit max time, ms
#define T_REP_MIN 10  //repeat pulse min delay, ms
#define T_REP_MAX 12  //repeat pulse max delay, ms
#define T_DAT_MIN 12  //data pulse min delay, ms
#define T_DAT_MAX 15  //data pulse max delay, ms

struct NECStruct //  NEC
{
  uint16_t NECCust;
  uint8_t NECpData;
  uint8_t NECnData;
};

#define PRE_NEC 1024 //timer prescaler
#define TICK (1E3 * PRE_NEC / F_CLK) //timer tick, ms
#define MS2TICK(ms) ((uint8_t)((ms) / TICK)) //ms to timer ticks
#define IS_T(t, min, max) (((t) >= MS2TICK(min)) && ((t) <= MS2TICK(max)))

enum { ST_IDLE, ST_WAIT, ST_DATA, ST_DONE }; //machine states

//----------------------------------------------------------------------------
//----------------------------  TRemote: --------------------------------
//----------------------------------------------------------------------------

//------------------------  : ---------------------------

uint8_t volatile  TRemote::State;
uint16_t volatile TRemote::System;
uint8_t volatile  TRemote::Command;

//----------------------------- : ---------------------------------

TRemote::TRemote(uint16_t sys)
{
  Pin_LED.DirOut(PIN_LOW);         //  
  Pin_IR.DirIn(PIN_PULL);          //  
  SysUsed = sys;                   //  
  Command = REM_NO;                //  
  State = ST_IDLE;                 // 
  TCCR2B = (1 << CS22) | (1 << CS21) | (1 << CS20); // CK/1024
  EICRA = 0;                       // INT0   
  EIFR = (1 << INTF0);             //  
  EIMSK |= (1 << INT0);            // INT0
  SMCR = (1 << SM1) | (1 << SE);   //Power Down enable
}

//-------------     : --------------

#pragma vector = INT0_vect
__interrupt void InputIR(void)
{
  static union
  {
    uint32_t RxCode;               // 
    NECStruct NECBytes;            //  
  };
  static uint8_t BitCount;

  uint8_t tm = TCNT2;              // 
  TCNT2 = 0;                       // 

  switch(TRemote::State)
  {
  // :
  case ST_IDLE:
    TIFR2 = (1 << TOV2);           //  
    TIMSK2 |= (1 << TOIE2);        //   2
    EICRA = (1 << ISC01);          // INT0  
    TRemote::Pin_LED = 1;          // 
    SMCR = 0;                      //sleep disable
    TRemote::State = ST_WAIT;
    break;
  // :
  case ST_WAIT:
  //   :
    if(IS_T(tm, T_DAT_MIN, T_DAT_MAX))
    {
      BitCount = 0;
      NECBytes.NECpData = REM_NO;
      TRemote::State = ST_DATA;
      break;
    }
    //  :
    if(IS_T(tm, T_REP_MIN, T_REP_MAX))
    {
      //  :
      if(NECBytes.NECpData == (uint8_t)~NECBytes.NECnData)
      {
        TRemote::System  = NECBytes.NECCust;
        TRemote::Command = NECBytes.NECpData;
        TRemote::State = ST_DONE;
        break;
      }
      TRemote::State = ST_IDLE;
      break;
    }
    NECBytes.NECpData = REM_NO;
    TRemote::State = ST_IDLE;
    break;
  // :
  case ST_DATA:
    if(IS_T(tm, T_ZER_MIN, T_ZER_MAX)) RxCode = RxCode & ((1UL << 31) - 1);
      else if(IS_T(tm, T_ONE_MIN, T_ONE_MAX)) RxCode = RxCode | (1UL << 31);
        else { TRemote::State = ST_IDLE; break; }
    if(++BitCount < NEC_BITS) RxCode = RxCode >> 1;
    else
    {
      if(NECBytes.NECpData == (uint8_t)~NECBytes.NECnData)
      {
        TRemote::System  = NECBytes.NECCust;
        TRemote::Command = NECBytes.NECpData;
        TRemote::State = ST_DONE;
        break;
      }
      TRemote::State = ST_IDLE;
    }
    break;
  }
}

//------------------    2: ------------------------

#pragma vector = TIMER2_OVF_vect
__interrupt void TimerIR(void)
{
  TRemote::State = ST_IDLE;
  TIMSK2 &= ~(1 << TOIE2);         //   2
  TRemote::Pin_LED = 0;            // 
  EICRA = 0;                       // INT0   
  SMCR = (1 << SM1) | (1 << SE);   //Power Down enable
}

//------------------------   : ------------------------------

uint8_t TRemote::GetCom(void)
{
  uint8_t c = REM_NO;
  if((State == ST_DONE) &&
     (System == SysUsed))
  {
    State = ST_IDLE;
    c = Command;
  }
  return(c);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
